/*
 * Copyright (C) 2020-2021  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/iface/report_manager_service.hh"
#include "src/iface/tools/tools.hh"

#include "src/iface/tools/common_types_wrappers.hh"
#include "src/iface/tools/generic_wrappers.hh"
#include "src/iface/tools/generic_unwrappers.hh"

#include "src/impl/manage_reports.hh"

#include "liblog/liblog.hh"

#include "libpg/pg_exception.hh"

namespace Fred {
namespace DbReport {
namespace Iface {

namespace {

decltype(auto) unwrap(const Api::ManageReportsRequest& api_request)
{
    Impl::Lib::ManageReportsRequest request;
    Tools::unwrap(api_request.sql(), request.sql);
    return request;
}

void wrap(const Impl::Lib::ManageReportsReply::Data& src, Api::ManageReportsReply::Data& dst)
{
    Tools::wrap(src.created, dst.mutable_created());
    Tools::wrap(src.deleted, dst.mutable_deleted());
}

void wrap(const Impl::Lib::ManageReportsReply& src, Api::ManageReportsReply& dst)
{
    wrap(src.data, *dst.mutable_data());
}

}//namespace Fred::DbReport::Iface::{anonymous}

grpc::Status ReportManagerService::manage_reports(
        grpc::ServerContext*,
        const Api::ManageReportsRequest* api_request,
        Api::ManageReportsReply* reply)
{
    LIBLOG_SET_CONTEXT(LogCtx, log_ctx, "ReportManager", __func__);
    try
    {
        const auto request = unwrap(*api_request);
        const auto result = Impl::manage_reports(request);
        wrap(result, *reply);
        return grpc::Status::OK;
    }
    catch (const Impl::NotSqlCommand& e)
    {
        return Tools::on_invalid_argument(e);
    }
    catch (const Impl::ManageReportsRequestSqlError& e)
    {
        return Tools::on_invalid_argument(e);
    }
    catch (const LibPg::ExecFailure& e)
    {
        return Tools::on_sql_error(e);
    }
    catch (const std::exception& e)
    {
        return Tools::on_std_exception(e);
    }
    catch (...)
    {
        return Tools::on_unknown_exception();
    }
}

}//namespace Fred::DbReport::Iface
}//namespace Fred::DbReport
}//namespace Fred
