/*
 * Copyright (C) 2020-2021  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/iface/report_user_service.hh"
#include "src/iface/tools/tools.hh"
#include "src/iface/tools/common_types_wrappers.hh"
#include "src/iface/tools/generic_wrappers.hh"
#include "src/iface/tools/generic_unwrappers.hh"
#include "src/iface/tools/run_report_unwrappers.hh"

#include "src/impl/exceptions.hh"
#include "src/impl/list_reports.hh"
#include "src/impl/run_report.hh"

#include "liblog/liblog.hh"

#include "libpg/pg_exception.hh"

#include <algorithm>
#include <cmath>

namespace Fred {
namespace DbReport {
namespace Iface {

namespace {

void append_rows(std::vector<Impl::Table::JsonEncodedFields>::const_iterator begin, std::vector<Impl::Table::JsonEncodedFields>::const_iterator end, Api::Table& table)
{
    table.mutable_rows()->Reserve(static_cast<int>(table.rows().size() + (end - begin)));
    std::for_each(begin, end, [&](const auto& row) { Tools::wrap(row, table.mutable_rows()->Add()); });
}

}//namespace Fred::DbReport::Iface::{anonymous}

grpc::Status ReportUserService::run_report(
        grpc::ServerContext*,
        const Api::RunReportRequest* api_request,
        grpc::ServerWriter<Api::RunReportReply>* writer)
{
    LIBLOG_SET_CONTEXT(LogCtx, log_ctx, "ReportUser", __func__);
    try
    {
        const auto request = [](const auto& api_request)
        {
            Impl::Lib::RunReportRequest impl_request;
            Tools::unwrap(api_request, impl_request);
            return impl_request;
        }(*api_request);
        const auto result = Impl::run_report(request);
        const auto total_rows = static_cast<int>(result.table.rows.size());
        const auto make_first_chunk = [&](Api::Table& first_chunk)
        {
            static constexpr int first_chunk_size_max = 256;
            const auto first_chunk_size = std::min(total_rows, first_chunk_size_max);
            Tools::wrap(result.table.field_descriptions, first_chunk.mutable_header()->mutable_field_descriptions());
            append_rows(std::begin(result.table.rows), std::begin(result.table.rows) + first_chunk_size, first_chunk);
            const bool just_one_chunk = first_chunk_size == total_rows;
            if (!just_one_chunk)
            {
                const auto first_chunk_memory_size = first_chunk.rows().SpaceUsedExcludingSelf();
                static constexpr long long kilo_byte = 1024;
                static constexpr auto mega_byte = 1024 * kilo_byte;
                static constexpr auto message_size_max = 1 * mega_byte;
                const auto chunk_size = static_cast<int>(std::lrint(double(first_chunk_size * message_size_max) / first_chunk_memory_size));
                if (chunk_size < first_chunk_size)
                {
                    throw std::runtime_error{"data too long"};
                }
                const auto current_chunk_size = std::min(chunk_size, total_rows);
                append_rows(std::begin(result.table.rows) + first_chunk_size, std::begin(result.table.rows) + current_chunk_size, first_chunk);
            }
        };
        Api::RunReportReply reply;
        auto& data = *reply.mutable_data();
        make_first_chunk(*data.mutable_table());
        writer->Write(reply);
        const auto first_chunk_size = data.table().rows().size();
        data.mutable_table()->clear_header();
        auto current_top = first_chunk_size;
        while (current_top < total_rows)
        {
            const auto next_top = (total_rows - current_top) <= first_chunk_size ? total_rows : current_top + first_chunk_size;
            data.mutable_table()->clear_rows();
            auto& current_chunk = *data.mutable_table();
            append_rows(std::begin(result.table.rows) + current_top, std::begin(result.table.rows) + next_top, current_chunk);
            writer->Write(reply);
        }
        return grpc::Status::OK;
    }
    catch (const LibPg::ExecFailure& e)
    {
        return Tools::on_sql_error(e);
    }
    catch (const Impl::InvalidArgument& e)
    {
        return Tools::on_invalid_argument(e);
    }
    catch (const std::exception& e)
    {
        return Tools::on_std_exception(e);
    }
    catch (...)
    {
        return Tools::on_unknown_exception();
    }
}

grpc::Status ReportUserService::list_reports(
        grpc::ServerContext*,
        const google::protobuf::Empty*,
        Api::ListReportsReply* reply)
{
    LIBLOG_SET_CONTEXT(LogCtx, log_ctx, "ReportUser", __func__);
    try
    {
        const auto result = Impl::list_reports();
        Tools::wrap(result.data.reports, reply->mutable_data()->mutable_reports());
        return grpc::Status::OK;
    }
    catch (const Impl::OverloadedReportName& e)
    {
        return Tools::on_failed_precondition(e);
    }
    catch (const LibPg::ExecFailure& e)
    {
        return Tools::on_sql_error(e);
    }
    catch (const std::exception& e)
    {
        return Tools::on_std_exception(e);
    }
    catch (...)
    {
        return Tools::on_unknown_exception();
    }
}

}//namespace Fred::DbReport::Iface
}//namespace Fred::DbReport
}//namespace Fred
