/*
 * Copyright (C) 2020-2021  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/impl/lib/list_reports.hh"

#include "liblog/liblog.hh"

#include "libpg/pg_ro_transaction.hh"
#include "libpg/query.hh"

#include <algorithm>
#include <map>
#include <sstream>
#include <stdexcept>
#include <utility>
#include <vector>


namespace Fred {
namespace DbReport {
namespace Impl {
namespace Lib {

namespace  {

struct Field
{
    using SpecificName = StrongType<std::string, struct SpecificNameTag_, Skill::Comparable>;
    using RoutineName = StrongType<std::string, struct RoutineNameTag_, Skill::Comparable>;
    using OrdinalPosition = StrongType<int, struct OrdinalPositionTag_, Skill::Countable>;
    using ParameterMode = StrongType<std::string, struct ParameterModeTag_, Skill::Comparable>;
    using ParameterName = StrongType<std::string, struct ParameterNameTag_>;
    using DataType = StrongType<std::string, struct DataTypeTag_>;
};

auto list_all_reports_query(const Schema& schema)
{
    return LibPg::make_query("SELECT") << LibPg::item<Field::SpecificName>() << "specific_name" <<
                                          LibPg::item<Field::RoutineName>() << "routine_name" <<
                                          LibPg::nullable_item<Field::OrdinalPosition>() << "ordinal_position" <<
                                          LibPg::nullable_item<Field::ParameterMode>() << "parameter_mode" <<
                                          LibPg::nullable_item<Field::ParameterName>() << "parameter_name" <<
                                          LibPg::nullable_item<Field::DataType>() << "data_type "
                             "FROM " << *schema << "._list_all_reports_()";
}

using Row = decltype(*exec(std::declval<const LibPg::PgRoTransaction&>(), list_all_reports_query(std::declval<const Schema&>())).begin());

struct FncInParameter
{
    Optional<Field::ParameterName> name;
    Field::DataType type;
};

struct FncOutParameter
{
    Optional<Field::ParameterName> name;
    Field::DataType type;
};

struct FncHeader
{
    Field::RoutineName fnc_name;
    std::vector<FncInParameter> in;
    std::vector<FncOutParameter> out;
};

using Functions = std::map<Field::SpecificName, FncHeader>;

void add_function(const Row& row, Functions& functions)
{
    auto& header = functions[row.get<Field::SpecificName>()];
    if (row.is_null<Field::OrdinalPosition>() ||
        (row.get_nullable<Field::OrdinalPosition>() == Field::OrdinalPosition{1}))
    {
        header.fnc_name = row.get<Field::RoutineName>();
    }
    else if (header.fnc_name != row.get<Field::RoutineName>())
    {
        throw std::logic_error{"routine name has to be the same for each parameter of one specific name"};
    }
    if (row.is_null<Field::OrdinalPosition>())
    {
        if (header.in.empty() && header.out.empty())
        {
            return;
        }
        throw std::logic_error{"ordinal position is NULL when routine does not have any parameters"};
    }
    if (row.is_null<Field::DataType>())
    {
        throw std::logic_error{"each parameter has to have data type specified"};
    }
    if (row.is_null<Field::ParameterMode>())
    {
        throw std::logic_error{"each parameter has to have mode specified"};
    }
    if (row.get_nullable<Field::ParameterMode>() == Field::ParameterMode{"IN"})
    {
        if (!header.out.empty())
        {
            throw std::logic_error{"input parameters have to be first"};
        }
        FncInParameter parameter;
        if (!row.is_null<Field::ParameterName>())
        {
            parameter.name = row.get_nullable<Field::ParameterName>();
        }
        parameter.type = row.get_nullable<Field::DataType>();
        header.in.emplace_back(std::move(parameter));
    }
    else if (row.get_nullable<Field::ParameterMode>() == Field::ParameterMode{"OUT"})
    {
        FncOutParameter parameter;
        if (!row.is_null<Field::ParameterName>())
        {
            parameter.name = row.get_nullable<Field::ParameterName>();
        }
        parameter.type = row.get_nullable<Field::DataType>();
        header.out.emplace_back(std::move(parameter));
    }
    else
    {
        throw std::runtime_error{"unsupported parameter mode"};
    }
    const auto ordinal_position = *row.get_nullable<Field::OrdinalPosition>();
    if (ordinal_position != static_cast<int>(header.in.size() + header.out.size()))
    {
        throw std::logic_error{"ordinal positions must be a continuous sequence starting with 1"};
    }
}

decltype(auto) get_functions(const LibPg::PgTransaction& tx, const Schema& schema)
{
    const auto db_res = exec(tx, list_all_reports_query(schema));
    Functions functions;
    std::for_each(db_res.begin(), db_res.end(), [&](Row row) { add_function(row, functions); });
    return functions;
}

decltype(auto) in_parameter_to_description(const FncInParameter& parameter)
{
    InputParameterDescription desc;
    if (parameter.name != Field::ParameterName::nullopt)
    {
        desc.name = ParameterName{**parameter.name};
    }
    desc.data_type = DataType{*parameter.type};
    return desc;
}

decltype(auto) out_parameter_to_description(const FncOutParameter& parameter)
{
    OutputParameterDescription desc;
    if (parameter.name != Field::ParameterName::nullopt)
    {
        desc.name = ParameterName{**parameter.name};
    }
    desc.data_type = DataType{*parameter.type};
    return desc;
}

decltype(auto) function_to_info(const FncHeader& fnc_header)
{
    ReportInfo info;
    info.name = ReportName{*fnc_header.fnc_name};
    std::transform(fnc_header.in.cbegin(), fnc_header.in.cend(), std::back_inserter(info.input_parameters), in_parameter_to_description);
    std::transform(fnc_header.out.cbegin(), fnc_header.out.cend(), std::back_inserter(info.output_parameters), out_parameter_to_description);
    return info;
}

decltype(auto) function_to_list_value(const Functions::value_type& fnc)
{
    ReportsList::value_type info{ReportId{*fnc.first}, function_to_info(fnc.second)};
    return info;
}

void out_input_param(const InputParameterDescription& param, bool& delimiter, std::ostream& out)
{
    if (delimiter)
    {
        out << ", ";
    }
    else
    {
        delimiter = true;
    }
    out << *param.data_type;
    if (!param.name->empty())
    {
        out << " \"" << *param.name << "\"";
    }
}

void out_output_param(const OutputParameterDescription& param, bool& delimiter, std::ostream& out)
{
    if (delimiter)
    {
        out << ", ";
    }
    else
    {
        delimiter = true;
    }
    out << *param.data_type;
    if (!param.name->empty())
    {
        out << " \"" << *param.name << "\"";
    }
}

decltype(auto) report_to_str(const ReportInfo& report)
{
    std::ostringstream out;
    out << *report.name << "(";
    bool delimiter = false;
    std::for_each(report.input_parameters.begin(), report.input_parameters.end(), [&](const auto& param) { out_input_param(param, delimiter, out); });
    out << ") -> (";
    delimiter = false;
    std::for_each(report.output_parameters.begin(), report.output_parameters.end(), [&](const auto& param) { out_output_param(param, delimiter, out); });
    out << ")";
    return std::move(out).str();
}

void log_report(const ReportsList::value_type& report)
{
    LIBLOG_TRACE("{}", report_to_str(report.second));
}

}//namespace Fred::DbReport::Impl::Lib::{anonymous}

ReportsList get_reports_list(const LibPg::PgTransaction& tx, const Schema& schema)
{
    ReportsList result{};
    const auto functions = get_functions(tx, schema);
    std::transform(functions.cbegin(), functions.cend(), std::inserter(result, result.begin()), function_to_list_value);
    std::for_each(result.begin(), result.end(), log_report);
    return result;
}

}//namespace Fred::DbReport::Impl::Lib
}//namespace Fred::DbReport::Impl
}//namespace Fred::DbReport
}//namespace Fred
