/*
 * Copyright (C) 2020-2021  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/impl/lib/manage_reports.hh"
#include "src/impl/lib/list_reports.hh"
#include "src/impl/lib/list_reports_reply.hh"

#include "libpg/query.hh"

namespace Fred {
namespace DbReport {
namespace Impl {
namespace Lib {

namespace  {

decltype(auto) exists(const ReportId& report_id, const ReportsList& list)
{
    return list.find(report_id) != list.end();
}

void add_deleted(const ReportsList& before, const ReportsList& after, std::vector<ReportName>& deleted)
{
    std::for_each(
            std::begin(before),
            std::end(before),
            [&](const ReportsList::value_type& info_before)
            {
                if (!exists(info_before.first, after))
                {
                    deleted.push_back(info_before.second.name);
                }
            });
}

void add_created(const ReportsList& before, const ReportsList& after, std::vector<ReportInfo>& created)
{
    std::for_each(
            std::begin(after),
            std::end(after),
            [&](const ReportsList::value_type& info_after)
            {
                if (!exists(info_after.first, before))
                {
                    created.push_back(info_after.second);
                }
            });
}

void check_uniqueness(const ListReportsReply::Data::Reports& list, const ReportInfo& info)
{
    if (0 < list.count(info))
    {
        struct UniqueConstraintViolation : OverloadedReportName
        {
            const char* what() const noexcept override { return "function name overloading not allowed"; }
        };
        throw UniqueConstraintViolation{};
    }
}

void uniqueness_required(const ReportsList& reports_list)
{
    ListReportsReply::Data::Reports unique_reports;
    std::transform(
            std::begin(reports_list),
            std::end(reports_list),
            std::inserter(unique_reports, unique_reports.begin()),
            [&](const auto& list_value)
            {
                check_uniqueness(unique_reports, list_value.second);
                return list_value.second;
            });
}

}//namespace Fred::DbReport::Impl::Lib::{anonymous}

ManageReportsReply manage_reports(const LibPg::PgRwTransaction& tx, const Schema& schema, const std::string& sql)
{
    const auto before = get_reports_list(tx, schema);
    LibPg::PgResultNoData{exec(tx, sql)};
    const auto after = get_reports_list(tx, schema);
    uniqueness_required(after);
    ManageReportsReply result{};
    add_deleted(before, after, result.data.deleted);
    add_created(before, after, result.data.created);
    return result;
}

}//namespace Fred::DbReport::Impl::Lib
}//namespace Fred::DbReport::Impl
}//namespace Fred::DbReport
}//namespace Fred
