/*
 * Copyright (C) 2020-2021  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/impl/lib/run_report.hh"
#include "src/impl/lib/list_reports.hh"

#include "libpg/query.hh"
#include "libpg/extendable_query_arguments.hh"

#include <algorithm>
#include <set>
#include <stdexcept>
#include <string>
#include <tuple>
#include <utility>
#include <vector>

namespace Fred {
namespace DbReport {
namespace Impl {
namespace Lib {

namespace  {

ReportInfo get_report(
        const LibPg::PgRoTransaction& tx,
        const Schema& schema,
        const ReportName& report_name)
{
    const auto reports = get_reports_list(tx, schema);
    const auto report_itr = std::find_if(
            std::begin(reports),
            std::end(reports),
            [&](const ReportsList::value_type& report) { return report.second.name == report_name; });
    if (report_itr == std::end(reports))
    {
        throw std::runtime_error{"database report does not exist"};
    }
    return report_itr->second;
}

decltype(auto) get_report_items(const std::vector<OutputParameterDescription>& output_fields)
{
    std::string report_items{};
    bool use_delimiter = false;
    std::for_each(
        std::begin(output_fields),
        std::end(output_fields),
        [&](const OutputParameterDescription& field)
        {
            if (!use_delimiter)
            {
                use_delimiter = true;
                report_items = "\"" + *field.name + "\"";
            }
            else
            {
                report_items += ",\"" + *field.name + "\"";
            }
        });
    return report_items;
}

decltype(auto) get_report_header(const std::vector<OutputParameterDescription>& output_fields)
{
    std::vector<Table::FieldDescription> header;
    std::transform(
        std::begin(output_fields),
        std::end(output_fields),
        std::back_inserter(header),
        [&](const OutputParameterDescription& field)
        {
            Table::FieldDescription header_field;
            header_field.data_type = field.data_type;
            header_field.name = *field.name;
            return header_field;
        });
    return header;
}

void append_report_argument(
        const Schema& schema,
        const JsonEncodedValue& argument,
        const DataType& data_type,
        std::string& sql,
        LibPg::ExtendableQueryArguments& query_arguments,
        bool use_delimiter)
{
    const auto delimiter = use_delimiter ? std::string{","}
                                         : std::string{};
    const auto json = LibPg::QueryParameterRuntimeType{"JSON"};
    sql.append(delimiter + *schema + "._arg_to_text_(" + query_arguments.add(argument).as(json) + ")::" + *data_type);
}

decltype(auto) get_report_arguments(
        const Schema& schema,
        const std::vector<InputParameterDescription>& parameters,
        const std::vector<JsonEncodedValue>& arguments)
{
    const auto parameter_begin = std::begin(parameters);
    const auto parameter_end = std::end(parameters);
    auto parameter_itr = parameter_begin;
    std::string sql_part{};
    bool use_delimiter = false;
    LibPg::ExtendableQueryArguments query_arguments;
    std::for_each(
            std::begin(arguments),
            std::end(arguments),
            [&](const JsonEncodedValue& argument)
            {
                if (parameter_itr == parameter_end)
                {
                    throw std::runtime_error{"too much arguments"};
                }
                append_report_argument(
                        schema,
                        argument,
                        parameter_itr->data_type,
                        sql_part,
                        query_arguments,
                        use_delimiter);
                use_delimiter = true;
                ++parameter_itr;
            });
    return std::make_tuple(std::move(sql_part), std::move(query_arguments));
}

decltype(auto) make_row(LibPg::PgResult::Row dbrow)
{
    Table::JsonEncodedFields row;
    row.fields = dbrow.at(LibPg::ColumnIndex{0}).as<std::string>();
    return row;
}

}//namespace Fred::DbReport::Impl::Lib::{anonymous}

RunReportReply run_report(const LibPg::PgRoTransaction& tx, const Schema& schema, const RunReportRequest& request)
{
    RunReportReply reply;
    const auto dbres = [&](std::vector<Table::FieldDescription>& field_descriptions)
    {
        const auto report = get_report(tx, schema, request.report_name);
        const auto report_items = get_report_items(report.output_parameters);
        field_descriptions = get_report_header(report.output_parameters);
        const auto report_arguments = get_report_arguments(schema, report.input_parameters, request.arguments);
        const std::string sql =
                "SELECT json_build_array(" + report_items + ") "
                "FROM " + *schema + "." + *report.name + "(" + std::get<std::string>(report_arguments) + ")";
        return LibPg::PgResultTuples{exec(tx, sql, std::get<LibPg::ExtendableQueryArguments>(report_arguments))};
    }(reply.table.field_descriptions);
    std::transform(std::begin(dbres), std::end(dbres), std::back_inserter(reply.table.rows), make_row);
    return reply;
}

void check_report_arguments(const LibPg::PgRoTransaction& tx, const Schema& schema, const RunReportRequest& request)
{
    const auto report = get_report(tx, schema, request.report_name);
    if (!report.input_parameters.empty())
    {
        const auto report_arguments = get_report_arguments(schema, report.input_parameters, request.arguments);
        const std::string sql = "SELECT " + std::get<std::string>(report_arguments);
        LibPg::PgResultTuples{exec(tx, sql, std::get<LibPg::ExtendableQueryArguments>(report_arguments))};
    }
}

}//namespace Fred::DbReport::Impl::Lib
}//namespace Fred::DbReport::Impl
}//namespace Fred::DbReport
}//namespace Fred
