/*
 * Copyright (C) 2020  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/impl/list_reports.hh"
#include "src/impl/lib/list_reports.hh"

#include "src/cfg.hh"

#include "src/util/util.hh"

#include "liblog/liblog.hh"

#include "libpg/query.hh"

#include <algorithm>

namespace Fred {
namespace DbReport {
namespace Impl {

namespace {

decltype(auto) get_reports_list(const Lib::Schema& schema)
{
    try
    {
        auto tx = Util::make_ro_transaction();
        auto result = Lib::get_reports_list(tx, schema);
        commit(std::move(tx));
        return result;
    }
    catch (const LibPg::ExecFailure& e)
    {
        LIBLOG_WARNING("SQL exec failure: {}", e.what());
        throw;
    }
    catch (const LibPg::ConnectFailure& e)
    {
        LIBLOG_WARNING("DB connect failure: {}", e.what());
        throw;
    }
    catch (const std::exception& e)
    {
        LIBLOG_WARNING("std::exception caught: {}", e.what());
        throw;
    }
}

void uniqueness_required(const Lib::ListReportsReply::Data::Reports& list, const ReportInfo& info)
{
    if (0 < list.count(info))
    {
        struct UniqueConstraintViolation : OverloadedReportName
        {
            const char* what() const noexcept override { return "function name overloading not allowed"; }
        };
        LIBLOG_WARNING("overloaded report name detected");
        throw UniqueConstraintViolation{};
    }
}

}//namespace Fred::DbReport::Impl::{anonymous}

Lib::ListReportsReply list_reports()
{
    LIBLOG_DEBUG("enter");
    const auto reports_list = get_reports_list(Lib::Schema{Cfg::Options::get().database.schema});
    Lib::ListReportsReply result;
    std::transform(
            std::begin(reports_list),
            std::end(reports_list),
            std::inserter(result.data.reports, result.data.reports.begin()),
            [&](const auto& list_value)
            {
                uniqueness_required(result.data.reports, list_value.second);
                return list_value.second;
            });
    LIBLOG_DEBUG("successfully done");
    return result;
}

}//namespace Fred::DbReport::Impl
}//namespace Fred::DbReport
}//namespace Fred
