/*
 * Copyright (C) 2020-2021  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/impl/manage_reports.hh"
#include "src/impl/lib/list_reports.hh"
#include "src/impl/lib/manage_reports.hh"

#include "src/cfg.hh"

#include "src/util/util.hh"

#include "liblog/liblog.hh"

#include "libpg/query.hh"

#include <algorithm>
#include <exception>

namespace Fred {
namespace DbReport {
namespace Impl {

namespace  {

bool is_same(const Lib::ReportsList& lhs, const Lib::ReportsList& rhs)
{
    return lhs == rhs;
}

decltype(auto) get_reports(const Lib::Schema& schema)
{
    auto tx = Util::make_ro_transaction();
    auto result = Lib::get_reports_list(tx, schema);
    commit(std::move(tx));
    return result;
}

template <typename Base>
[[noreturn]] void rethrow_as(const std::exception& message_producer)
{
    struct Exception : Base
    {
        explicit Exception(const std::exception& message_producer)
            : Base{},
              msg{message_producer.what()}
        { }
        const char* what() const noexcept override
        {
            return msg.c_str();
        }
        std::string msg;
    };
    throw Exception{message_producer};
}

template <typename Base>
[[noreturn]] void raise(const char* message)
{
    struct Exception : Base
    {
        explicit constexpr Exception(const char* message)
            : Base{},
              message{message}
        { }
        const char* what() const noexcept { return message; }
        const char* message;
    };
    throw Exception{message};
}

}//namespace Fred::DbReport::Impl::{anonymous}

Lib::ManageReportsReply manage_reports(const Lib::ManageReportsRequest& request)
{
    LIBLOG_DEBUG("enter");
    const auto schema = Lib::Schema{Cfg::Options::get().database.schema};
    auto tx = Util::make_rw_transaction();
    try
    {
        const auto before = get_reports(schema);
        auto result = [&]()
        {
            try
            {
                return Lib::manage_reports(tx, schema, request.sql);
            }
            catch (const LibPg::ConnectFailure&)
            {
                throw;
            }
            catch (const LibPg::ExecFailure& e)
            {
                rethrow_as<ManageReportsRequestSqlError>(e);
            }
            catch (const LibPg::NoConnectionAvailable&)
            {
                throw;
            }
            catch (const LibPg::PgException& e)
            {
                rethrow_as<NotSqlCommand>(e);
            }
            catch (const Lib::OverloadedReportName& e)
            {
                rethrow_as<ManageReportsRequestSqlError>(e);
            }
        }();
        const auto after = get_reports(schema);
        if (!is_same(before, after))
        {
            rollback(std::move(tx));
            raise<RaceConditionDetected>("reports changed outside transaction");
        }
        commit(std::move(tx));
        LIBLOG_DEBUG("successfully done");
        return result;
    }
    catch (const ManageReportsRequestSqlError&)
    {
        throw;
    }
    catch (const LibPg::ConnectFailure& e)
    {
        LIBLOG_WARNING("DB connect failure: {}", e.what());
        throw;
    }
    catch (const LibPg::ExecFailure& e)
    {
        LIBLOG_WARNING("SQL exec failure: {}", e.what());
        throw;
    }
    catch (const LibPg::PgException& e)
    {
        LIBLOG_WARNING("PgException: {}", e.what());
        throw;
    }
    catch (const RaceConditionDetected& e)
    {
        LIBLOG_INFO("race condition detected: {}", e.what());
        throw;
    }
    catch (const std::exception& e)
    {
        LIBLOG_WARNING("std::exception caught: {}", e.what());
        throw;
    }
}

}//namespace Fred::DbReport::Impl
}//namespace Fred::DbReport
}//namespace Fred
