/*
 * Copyright (C) 2020-2021  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/impl/run_report.hh"
#include "src/impl/exceptions.hh"
#include "src/impl/lib/run_report.hh"

#include "src/cfg.hh"

#include "src/util/util.hh"

#include "liblog/liblog.hh"

#include "libpg/query.hh"
#include "libpg/sql_state/code.hh"

namespace Fred {
namespace DbReport {
namespace Impl {

Lib::RunReportReply run_report(const Lib::RunReportRequest& request)
{
    LIBLOG_DEBUG("enter");
    auto tx = Util::make_ro_transaction(Util::isolation_level_for_reports);
    try
    {
        auto result = Lib::run_report(tx, Lib::Schema{Cfg::Options::get().database.schema}, request);
        commit(std::move(tx));
        LIBLOG_DEBUG("successfully done");
        return result;
    }
    catch (const LibPg::ExecFailure& e)
    {
        LIBLOG_WARNING("SQL exec failure: {}", e.what());
        auto conn = rollback(std::move(tx));
        constexpr auto invalid_text_representation = LibPg::SqlState::data_exception_class.subclass<'P', '0', '2'>();
        if (e.error_code() == invalid_text_representation)
        {
            LibPg::PgRoTransaction tx{std::move(conn), Util::isolation_level_for_reports};
            try
            {
                Lib::check_report_arguments(tx, Lib::Schema{Cfg::Options::get().database.schema}, request);
                commit(std::move(tx));
                LIBLOG_INFO("Report arguments are OK");
            }
            catch (const LibPg::ExecFailure& e)
            {
                rollback(std::move(tx));
                if (e.error_code() == invalid_text_representation)
                {
                    class InvalidReportArguments : public InvalidArgument
                    {
                    public:
                        InvalidReportArguments(const char* msg) : msg_{msg} { }
                        const char* what() const noexcept override { return msg_.c_str(); }
                    private:
                        std::string msg_;
                    };
                    LIBLOG_WARNING("Report arguments are invalid: {}", e.what());
                    throw InvalidReportArguments{e.what()};
                }
            }
        }
        throw;
    }
    catch (const LibPg::ConnectFailure& e)
    {
        LIBLOG_WARNING("DB connect failure: {}", e.what());
        throw;
    }
    catch (const std::exception& e)
    {
        LIBLOG_WARNING("std::exception caught: {}", e.what());
        rollback(std::move(tx));
        throw;
    }
}

}//namespace Fred::DbReport::Impl
}//namespace Fred::DbReport
}//namespace Fred
