/*
 * Copyright (C) 2020-2021  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "setup.hh"

#include "src/impl/lib/list_reports.hh"
#include "src/impl/lib/manage_reports.hh"
#include "src/impl/lib/run_report.hh"

#include "src/cfg.hh"

#include <boost/test/unit_test.hpp>

#include <utility>

BOOST_AUTO_TEST_SUITE(TestManageReport)

namespace {

decltype(auto) drop_function(const Fred::DbReport::Impl::Lib::Schema& schema, const Fred::DbReport::Impl::ReportName& name, const std::string& parameters)
{
    const auto sql = "DROP FUNCTION IF EXISTS " + *schema + "." + *name + "(" + parameters + ");";
    auto tx = Test::make_rw_transaction();
    auto result = Fred::DbReport::Impl::Lib::manage_reports(tx, schema, sql);
    commit(std::move(tx));
    return result.data;
}

}//namespace {anonymous}

BOOST_AUTO_TEST_CASE(manage)
{
    const auto schema = Fred::DbReport::Impl::Lib::Schema{Cfg::Options::get().database.schema};
    const auto report_name = Fred::DbReport::Impl::ReportName{"test_domain_delete_candidates"};
    {
        drop_function(schema, report_name, "DATE,DATE");
        const auto result = [&]()
        {
            const auto sql =
                "CREATE OR REPLACE FUNCTION " + *schema + "." + *report_name + "(t0 DATE = CURRENT_DATE-'1MONTH'::INTERVAL, t1 DATE=CURRENT_DATE)\n"
                "RETURNS TABLE(\"date\" DATE,\n"
                "              \"delete candidates\" BIGINT) AS $$\n"
                "WITH t AS (SELECT t0+s.v AS d FROM generate_series(0,t1-t0) AS s(v))\n"
                "SELECT t.d,COUNT(DISTINCT obr.id)\n"
                "FROM t\n"
                "LEFT JOIN public.object_state os ON os.valid_from::DATE=t.d\n"
                "LEFT JOIN public.enum_object_states eos ON eos.id=os.state_id AND eos.name='deleteCandidate'\n"
                "LEFT JOIN public.object_registry obr ON obr.id=os.object_id AND obr.type=public.get_object_type_id('domain')\n"
                "GROUP BY 1\n"
                "ORDER BY 1;\n"
                "$$\n"
                "LANGUAGE SQL\n"
                "STABLE\n"
                "SECURITY DEFINER\n"
                "SET search_path=" + *schema + ",pg_temp;";
            auto tx = Test::make_rw_transaction();
            const auto result = Fred::DbReport::Impl::Lib::manage_reports(tx, schema, sql);
            commit(std::move(tx));
            return result.data;
        }();
        BOOST_REQUIRE(result.deleted.empty());
        BOOST_REQUIRE_EQUAL(result.created.size(), 1);
        BOOST_REQUIRE_EQUAL(*result.created.front().name, *report_name);
        const auto report_list = [&]()
        {
            auto tx = Test::make_ro_transaction();
            const auto result = Fred::DbReport::Impl::Lib::get_reports_list(tx, schema);
            commit(std::move(tx));
            return result;
        }();
        BOOST_CHECK_LE(1, report_list.size());
        const auto report = std::find_if(
                std::begin(report_list),
                std::end(report_list),
                [&](const auto& report) { return report.second.name == report_name; });
        const bool has_report = report != std::end(report_list);
        BOOST_CHECK(has_report);
        if (has_report)
        {
            Fred::DbReport::Impl::Lib::RunReportRequest request;
            request.report_name = report_name;
            request.arguments = {Fred::DbReport::Impl::JsonEncodedValue{"\"2018-02-24\""},
                                 Fred::DbReport::Impl::JsonEncodedValue{"\"2018-03-24\""}};
            auto tx = Test::make_ro_transaction(Test::isolation_level_for_reports);
            BOOST_CHECK_NO_THROW(Fred::DbReport::Impl::Lib::run_report(tx, schema, request));
            commit(std::move(tx));
        }
    }
    const auto result = drop_function(schema, report_name, "DATE,DATE");
    BOOST_REQUIRE_EQUAL(result.deleted.size(), 1);
    BOOST_REQUIRE_EQUAL(*result.deleted.front(), *report_name);
    BOOST_REQUIRE(result.created.empty());
}

BOOST_AUTO_TEST_SUITE_END()//TestManageReport
